#function to compute the ML tree for a given data set
#
#starts with two aberrations and adds the other aberrations in random order, maximizing the
#likelihood in each step by trying all possible edges as insertion point
#
# for each newly inserted edge, a local maximum of the likelihood is sought by global rearrangement.
# with the new method: q's as cond. frquencies
# if any p[j] is >1, delete corresponding edge, until all p<=1
# if a node has more than two children, the corresponding q's are computed numerically
MLtopology <- function(x, verbose=FALSE) {
  if (!is.data.frame(x) & !is.matrix(x))
    stop("'x' should be of type 'matrix'.")
  x <- as.matrix(x)
  if (!(all(x %in% c(0,1))))
    stop("'x' should be binary.")
  if (any(colSums(x) == 0))
    stop("Every event should occur in at least one case.")
  if (is.null(colnames(x)))
    stop("'x' should have column names.")
  #start with a randomly chosen aberration:
  used <- sample(ncol(x), 2)
  oldtree <- cbind(c(3,1), c(3,2), c(4,3))
  #the next ones (in random order):
  order <- sample((1:ncol(x))[-used])
 
  for (k in 1:length(order)) {
    #new indices for the new tree:
    oldtree[which(oldtree>length(used))] <- oldtree[which(oldtree>length(used))] + 1
    bestvalue <- -Inf
    used <- c(used, order[k])
  #0.given binary data matrix, write a new one with unique rows and frequencies freq
    dat <- x[,used]
    naber <- ncol(dat)
    dat <- cbind(dat, dat %*% 2^(0:(naber-1)))
    newdata <- dat[which(!duplicated(dat[,naber+1])),1:naber]
    uniq <- unique(dat[,naber+1])
    freq <- rep(NA, nrow(newdata))
    for (j in 1:length(freq))
      freq[j] <- length(which(dat[,naber+1] == uniq[j]))
    tree <- cbind(c(2*length(used),length(used)), c(2*length(used),oldtree[2,1]), c(oldtree[1,1], 2*length(used)), oldtree[,2:ncol(oldtree)])
    res <- MLparameters(newdata, tree, freq)
    besttree <- tree
    if (!is.na(res$totloglik))
      bestvalue <- res$totloglik
    for (edge in 2:ncol(oldtree)) {
      tree <- cbind(oldtree[,which(1:ncol(oldtree)<edge)], c(2*length(used),length(used)), c(2*length(used),oldtree[2,edge]), c(oldtree[1,edge], 2*length(used)), oldtree[,which(1:ncol(oldtree)>edge)])
      if (!is.na(res$totloglik) & res$totloglik > bestvalue) {
        besttree <- tree
        bestvalue <- res$totloglik
      }
    }
    oldtree <- besttree
    if (verbose == TRUE) {
      cat("adding", colnames(x)[order[k]], "\t")
      cat(bestvalue, "\n")
    }
    oldbestvalue <- bestvalue
    curedge <- which(besttree[2,]==length(used))
  #oldtree is the best tree resulting from inserting the new edge
  #
  #now perform rearrangements:
    repeat {
      flag <- 0
      for (rmedge in (1:ncol(oldtree))[-curedge]) {
        tree <- oldtree
        subtreeind <- c(subtree(tree[2,rmedge], tree), rmedge)
    #exclude the case where the removed tree is almost everything 
        if (ncol(tree) - length(subtreeind) > 2) { 
          subt <- as.matrix(tree[,subtreeind]) #the cut off tree 
          rest <- as.matrix(tree[,-subtreeind]) #the remainder of the old tree
          restparents <- match(rest[1,],rest[2,])
          sib <- which(rest[1,]==tree[1,rmedge])
          if (length(sib)==1) { #should be the case anyway; merge two edges 
            rest[1,sib] <- rest[1,restparents[sib]]
            rest <- as.matrix(rest[,-restparents[sib]])
          }
          for (insedge in 1:ncol(rest)) { # insert the cut off tree here
            insedgeintree <- which(oldtree[1,]==rest[1,insedge] & oldtree[2,]==rest[2,insedge])
            subtreeind2 <- subtree(rest[2,insedge], rest) # what if empty?
            subt2 <- as.matrix(rest[,subtreeind2]) #the part below the insertion edge
            rest2 <- rest[,-c(insedge, subtreeind2)]
            newtree <- cbind(subt, subt2, c(oldtree[1,rmedge], rest[2,insedge]), c(rest[1,insedge], oldtree[1,rmedge]), rest2)
            tree <- newtree
            res <- MLparameters(newdata, tree, freq)
          if (!is.na(res$totloglik) & res$totloglik - bestvalue > 1e-10) { #rearrangement led to a better tree
            besttree <- tree
            bestvalue <- res$totloglik
            curedge <- ncol(subt)
            flag <- 1
          }
          }
        }
      } #for rmedge
      if (flag==1) {
        if (verbose == TRUE)
          cat("change due to rearrangement\t", bestvalue, "improvement by", bestvalue-oldbestvalue, "\n")
        oldtree <- besttree
        oldbestvalue <- bestvalue
      } else {
        break
      }
    } #repeat
  }  #for (k in 1:length(order))
  totloglik <- bestvalue

#again compute the p's for the resulting tree:
  tree <- besttree
  res <- MLparameters(newdata, tree, freq)
  cat("resulting tree", res$totloglik, "\n")
#return the tree with the original order of the aberrations
  tr <- tree[2,]
  tr[which(tree[2,]<=naber)] <- used[tree[2,which(tree[2,]<=naber)]]
  tree[2,] <- tr
  p <- res$p
  totloglik <- res$totloglik
  output <- list(tree, p, totloglik)
  names(output) <- c("tree", "p", "totloglik")
  return(output)
}
